'----------------------------------------------------------------------------------------------------------------------
'
' In place Radix-2 Decimation In Time FFT
'
' Michael Rychlik. 2011-1-22
'
' This file is relesed under the terms of the MIT license. See below.
'
' This FFT was developed from the description by Douglas L. Jones at
' http://cnx.org/content/m12016/latest/.
' It is written as a direct implementation of the discussion and diagrams on that page
' with an emphasis on clarity and ease of understanding rather than speed.
'
' There are two implementations here. The initial version in Spin and a PASM version derived from it.
' Use #define SPIN_BUTTERFLIES or PASM_BUTTERFLIES to select them
'
' The PASM verson does a 1024 FFT in 40ms.
'
' WARNING: I have written this, my first ever FFT implememtation, as a challenge to myself to understand
' Cooley-Tookey from the maths to the code without peeking at the example code on the above page or any other.
' So be warned it may not be accurate.
'
' Credits:
'
'     A big thank you to Dave Hein for clearing up some issues during a great FFT debate on
'     the Parallax Inc Propller discussion forum:
'     http://forums.parallax.com/showthread.php?127306-Fourier-for-dummies-under-construction
'
'     Thanks to Lonesock for the multiply routines.
'
'     Another big thank you to Lonesock for suggesting and implementing the W0 optimization.
'
' Instructions:

'     0) Define PASM_BUTTERFLIES or SPIN_BUTTERFLIES, see below.
'     1) Application should reserve two 1024 arrays of LONGs for input and output data, x and y
'     2) Place input signal into the x array, max value 4095
'     3) Clear the x array.
'     4) Call buuterflies giving a command and addresses of the arrays as parameters.
'     5) The command is one or more of the following bits set:
'               CMD_DECIMATE   - Perform bit-reversal reordering on the data, results in x and y
'               CMD_BUTTERFLY  - Perform the actuall FFT butterfly calculations, results in x and y
'               CMD_MAGNITUDE  - Convert resulting x and y values to frequency magnitudes in x
'
'     6) For some applications it may be as well to write the input data directly into x LONG by LONG
'        in the correct bit-reversed order. Then drop the CMD_DECIMATE for the butterfly call.
'        This could move 10 percent or so of the processing time to the input COG.
'
' History:
'
' 2010-12-16    v0.1  Initial draft.
'
' 2010-12-18    v0.2  First working version.
'
' 2010-12-18    v0.3  Optimized butterfly from 4 multiplies to 3
'                     Reduced twiddle tables from longs to words
'
' 2010-12-21    v0.4  A few simple optimizations
'                     First PASM implementation.
'                     Split out test harness.
'
' 2010-12-28    v0.5  Added getsin and getcos fuctions, not taken into use yet.
'
' 2011-01-14    v1.0  Total rewrite of Spin butterflies method.
'                     Now uses pointers into data arrays rather than array indexing.
'                     Total rewrite of the PASM butterflies to match the above.
'                     This speeds the PASM butterflies by about 25%.
'                     Added copyright and MIT licence statemets.
'
' 2011-01-14    v1.1  Optimization by Lonesock. For each "flight" the first butterfly has "twiddle factor"
'                     wx = 1 (The cos part) and wy = 0 (The sine part). So we can avoid doing the
'                     multiplies in this case. This is only implemented in the PASM version so as to keep
'                     the original Spin code clean.
'
' 2011-01-22    v2.0  Further optimization by Lonesock for the case when the "twiddle factor" is wx=0, wy=-1
'                     Chopped out half of the wy table as it's thes ame as the second half of wx.
'                     Moved x and y buffers out of heater_fft.
'                     Added a mailbox interface to PASM.
'                     Added bit-reversal step as an option.
'                     Added magnitude calculation as an option.
'                     Thanks again to Lonesock for the faster square root routine.
'
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
'User optimization controls
#define PASM_BUTTERFLIES       'Set this for fast PASM FFT, about 30ms
'#define SPIN_BUTTERFLIES      'Or set this fro slow Spin FFT, about 1800ms!!

#define USE_FASTER_MULT        'Set this for faster multiply

'#define USE_FASTER_SQRT       'Set this for faster but much bigger square root.
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
CON
    'Specify size of FFT buffer here with length and log base 2 of the length.
    'N.B. Changing this will require changing the "twiddle factor" tables.
    '     and may also require changing the fixed point format (if going bigger)
    FFT_SIZE      = 1024
    LOG2_FFT_SIZE = 10

    CMD_DECIMATE  = %001
    CMD_BUTTERFLY = %010
    CMD_MAGNITUDE = %100
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
VAR
    long mailboxp
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
PUB start (mailp)
#ifdef PASM_BUTTERFLIES
    mailboxp := mailp
    LONG[mailboxp] := 0

    wx_ptr := @wx[0]
    wy_ptr := @wy[0]
    cognew (@bfly, mailp)     'Check error?
#endif
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
#ifdef SPIN_BUTTERFLIES
VAR
    long level
    long flight_max
    long flight
    long butterflySpan
    long butterfly_max
    long butterfly
    long temp
    long flightSkip
    long wSkip

    long b0x_ptr
    long b0y_ptr
    long b1x_ptr
    long b1y_ptr
    long wx_ptr
    long wy_ptr

    long a
    long b
    long c
    long d
    long k1
    long k2
    long k3
    long tx
    long ty
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
PUB  butterflies (cmd, bxp, byp)
    if cmd & CMD_DECIMATE                          'Data bit-reversal reordering required?
        decimate(bxp, byp)
    if cmd & CMD_BUTTERFLY                         'FFT butterfly required?
        bfly (bxp, byp)
    if cmd & CMD_MAGNITUDE                         'Convert to magnitude required?
        magnitude (bxp, byp)
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
PUB decimate (bxp, byp) | i, revi, tx1, ty1
'Radix-2 decimation in time.
'Moves every sample of bx and by to a postion given by
'reversing the bits of its original array index.
    repeat i from 0 to FFT_SIZE - 1
        revi := i >< LOG2_FFT_SIZE
        if i < revi
            tx1 := long[bxp + i * 4]
            ty1 := long[byp + i * 4]

            long[bxp + i * 4] := long[bxp + revi * 4]
            long[byp + i * 4] := long[byp + revi * 4]

            long[bxp + revi * 4] := tx1
            long[byp + revi * 4] := ty1
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
PUB magnitude (bxp, byp) | i, real, imag
    repeat i from 0 to (FFT_SIZE / 2)
        'Scale down by half FFT size, back to original signal input range
        real := long[bxp + i * 4] / (FFT_SIZE / 2)
        imag := long[byp + i * 4] / (FFT_SIZE / 2)

        'Frequency magnitude is square root of cos part sqaured plus sin part squared
        long[bxp + i * 4] := ^^((real * real) + (imag * imag))
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
'Apply FFT butterflies to N complex samples in x, in time decimated order !
'Resulting FFT is in x in the correct order.
PUB bfly (bxp, byp)
    flight_max := FFT_SIZE >> 1 ' / 2              'Initial number of flights in a level
    wSkip := FFT_SIZE                              'But we advance w pointer by 2 bytes per entry
    butterflySpan := 4                             'Span measured in bytes
    butterfly_max := 1                             '1 butterfly per flight initially
    flightSkip := 4                                'But we advance pointer by 4 bytes per butterfly

    'Loop through all the decimation levels
    repeat LOG2_FFT_SIZE
        b0x_ptr := bxp
        b0y_ptr := byp

        b1x_ptr := b0x_ptr + butterflySpan
        b1y_ptr := b0y_ptr + butterflySpan

        'Loop though all the flights in a level
        repeat flight_max
            wx_ptr := @wx
            wy_ptr := @wy

            'Loop through all the butterflies in a flight
            repeat butterfly_max
                'At last...the butterfly.
                '----------------------
                a := LONG[b1x_ptr]                 'Get X[b1]
                b := LONG[b1y_ptr]

                c := ~~WORD[wx_ptr]                'Get W[wIndex]
                d := ~~WORD[wy_ptr]

                k1 := (a * (c + d)) ~> 12 ' / 4096 'Somewhat optimized complex multiply
                k2 := (d * (a + b)) ~> 12 ' / 4096 '   T = X[b1] * W[wIndex]
                k3 := (c * (b - a)) ~> 12 ' / 4096
                tx := k1 - k2
                ty := k1 + k3

                k1 := LONG[b0x_ptr]                'bx[b0]
                k2 := LONG[b0y_ptr]                'by[b0]
                LONG[b1x_ptr] := k1 - tx           'X[b1] = X[b0] - T
                LONG[b1y_ptr] := k2 - ty

                LONG[b0x_ptr] := k1 + tx           'X[b0] = X[b0] + T
                LONG[b0y_ptr] := k2 + ty
                '---------------------

                b0x_ptr += 4                       'Advance to next butterfly in flight,
                b0y_ptr += 4                       'skiping 4 bytes for each.

                b1x_ptr += 4
                b1y_ptr += 4

                wx_ptr += wSkip                    'Advance to next w
                wy_ptr += wSkip

            b0x_ptr += flightSkip                  'Advance to first butterfly of next flight
            b0y_ptr += flightSkip
            b1x_ptr += flightSkip
            b1y_ptr += flightSkip

        butterflySpan <<= 1                        'On the next level butterflies are twice as wide
        flightSkip <<= 1                           'and so is the flight skip

        flight_max >>= 1                           'On the next level there are half as many flights
        wSkip >>= 1                                'And w's are half as far apart
        butterfly_max <<= 1                        'On the next level there are twice the butterflies per flight
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
#endif
#ifdef PASM_BUTTERFLIES
PUB butterflies(cmd, bxp, byp)
    LONG[mailboxp + 4] := bxp                          'Address of x buffer
    LONG[mailboxp + 8] := byp                          'Address of y buffer
    LONG[mailboxp + 0] := cmd                          'Do butterflies and/or decimation
    repeat while LONG[mailboxp + 0] <> 0
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
DAT
              org       0
bfly          mov       mb_ptr, par
              rdlong    command, mb_ptr wz                 'Wait for run command in mailbox
        if_z  jmp       #bfly

              add       mb_ptr, #4                         'Fetch x array address from mbox
              rdlong    bx_ptr, mb_ptr

              add       mb_ptr, #4                         'Fetch y array address from mbox
              rdlong    by_ptr, mb_ptr
              sub       mb_ptr, #8

              test      command, #CMD_DECIMATE wz          'Bit reversal required on data?
        if_z  jmp       #:no_decimate

'Radix-2 decimation in time. (The bit reversal satge)
'Moves every sample of bx to a postion given by reversing the bits of its original array index.
'This is a direct translation of the Spin demate above, original Spin code used as comments.
'N.B. Only the x array is bit-reversed it is up to the app to clear y.

              mov       c, fft_size_                       'repeat i from 0 to FFT_SIZE - 1
              mov       b, #0

:dloop        mov       a, b                               'revi := i >< LOG2_FFT_SIZE
              mov       rev_a, a
              rev       rev_a, #32 - LOG2_FFT_SIZE

              cmp       a, rev_a wc                        'if i < revi
        if_nc jmp       #:skip_rev

              shl       a, #2                              'Times 4 as we are reading longs
              shl       rev_a, #2

              mov       hub_ptr, bx_ptr                    'tx1 := long[bxp + i * 4]
              add       hub_ptr, a
              rdlong    tx, hub_ptr

              mov       hub_rev_ptr, bx_ptr                'long[bxp + i * 4] := long[bxp + revi * 4]
              add       hub_rev_ptr, rev_a
              rdlong    ty, hub_rev_ptr
              wrlong    ty, hub_ptr

              wrlong    tx, hub_rev_ptr                    'long[bxp + revi * 4] := tx1

:skip_rev     add       b, #1
              djnz      c, #:dloop

:no_decimate
              test      command, #CMD_BUTTERFLY wz         'Perform buterflies?
        if_z  jmp       #:no_butterfly

'Apply FFT butterflies to N complex samples in buffers bx and by, in time decimated order!
'Resulting FFT is produced in bx and by in the correct order.
'This is a direct translation from the Spin code above, original Spin code in comments.

              mov       flight_max, fft_size_              'flight_max := FFT_SIZE / 2
              sar       flight_max, #1
              mov       wSkip, fft_size_                   'wSkip := FFT_SIZE

              mov       butterflySpan, #4                  'butterflySpan := 4
              mov       butterfly_max, #1                  'butterfly_max := 1
              mov       flightSkip, #4                     'flightSkip := 4

              'Loop through all the decimation levels
              mov       level, #LOG2_FFT_SIZE              'level := LOG2_FFT_SIZE
:lloop                                                     'repeat
              mov       b0x_ptr, bx_ptr                    'b0x_ptr := @bx
              mov       b0y_ptr, by_ptr                    'b0y_ptr := @by

              mov       b1x_ptr, b0x_ptr                   'b1x_ptr := b0x_ptr + butterflySpan
              add       b1x_ptr, butterflySpan

              mov       b1y_ptr, b0y_ptr                   'b1y_ptr := b0y_ptr + butterflySpan
              add       b1y_ptr, butterflySpan

              'Loop though all the flights in a level
              mov       flight, flight_max                 'flight := flight_max
:floop                                                     'repeat
              mov       wx_ptr_, wx_ptr                    'wx_ptr := @wx
              mov       wy_ptr_, wy_ptr                    'wy_ptr := @wy

              'Loop through all the butterflies in a flight
              mov       butterfly, butterfly_max           'butterfly := butterfly_max

              'Do the initial pass optimization, when W = [1,0] we don't need to multiply.
              ' c = 1 (well, 4096/4096), d = 0
              mov       k2, #0                             'k2 := (d * (a + b)) / 4096
              rdlong    a, b1x_ptr                         'a := LONG[b1x_ptr]
              mov       k1, a                              'k1 := (a * (c + d)) / 4096
              neg       k3, a                              'k3 := (c * (b - a)) / 4096 
              rdlong    b, b1y_ptr                         'b := LONG[b1y_ptr]
              add       k3, b                              'k3 := (c * (b - a)) / 4096 (cont.)              
              jmp       #:continue_bloop
              
:bloop        ' repeat                                     'At last...the butterfly.
              rdlong    a, b1x_ptr                         'a := LONG[b1x_ptr]

              'Precompute the optimization for c=0, d=-1
              neg       k1, a                              'k1 := (a * (c + d)) / 4096
              neg       k2, a                              'k2 := (d * (a + b)) / 4096

              rdlong    b, b1y_ptr                         'b := LONG[b1y_ptr]

              'Precompute the optimization for c=0, d=-1
              sub       k2, b                              'k2 := (d * (a + b)) / 4096 (cont.)
              mov       k3, #0                             'k3 := (c * (b - a)) / 4096

              rdword    c, wx_ptr_ wz                      'c := ~~WORD[wx_ptr]

        if_z  jmp       #:continue_bloop                   ' if c==0, we already kave k1, k2, k3 calculated

              shl       c, #16
              rdword    d, wy_ptr_                         'd := ~~WORD[wy_ptr]
              sar       c, #16
              shl       d, #16
              sar       d, #16

              mov       m1, c                              'k1 := (a * (c + d)) / 4096
              add       m1, d
              mov       m2, a
              call      #mul
              mov       k1, m1
              sar       k1, #12

              mov       m1, a                              'k2 := (d * (a + b)) / 4096
              add       m1, b
              mov       m2, d
              call      #mul
              mov       k2, m1
              sar       k2, #12

              mov       m1, b                              'k3 := (c * (b - a)) / 4096
              sub       m1, a
              mov       m2, c
              call      #mul
              mov       k3, m1
              sar       k3, #12

:continue_bloop

              mov       tx, k1                             'tx := k1 - k2 (part I)
              mov       ty, k1                             'ty := k1 + k3 (part I)

              rdlong    k1, b0x_ptr                        'k1 := LONG[b0x_ptr]

              sub       tx, k2                             ' (part II) moved from above to take advantage of the hub wait times
              add       ty, k3                             ' ditto

              rdlong    k2, b0y_ptr                        'k2 := LONG[b0y_ptr]

              mov       a, k1                              'LONG[b1x_ptr] := k1 - tx
              sub       a, tx
              wrlong    a, b1x_ptr

              mov       a, k2                              'LONG[b1y_ptr] := k2 - ty
              sub       a, ty
              wrlong    a, b1y_ptr

              mov       a, k1                              'LONG[b0x_ptr] := k1 + tx
              add       a, tx
              wrlong    a, b0x_ptr

              mov       a, k2                              'LONG[b0y_ptr] := k2 + ty
              add       a, ty
              wrlong    a, b0y_ptr

              add       b0x_ptr, #4                        'b0x_ptr += 4
              add       b0y_ptr, #4                        'b0y_ptr += 4

              add       b1x_ptr, #4                        'b1x_ptr += 4
              add       b1y_ptr, #4                        'b1y_ptr += 4

              add       wx_ptr_, wSkip                     'wx_ptr += wSkip
              add       wy_ptr_, wSkip                     'wy_ptr += wSkip
              djnz      butterfly, #:bloop                 'while --butterfly <> 0

              add       b0x_ptr, flightSkip                'b0x_ptr += flightSkip
              add       b0y_ptr, flightSkip                'b0y_ptr += flightSkip
              add       b1x_ptr, flightSkip                'b1x_ptr += flightSkip
              add       b1y_ptr, flightSkip                'b1y_ptr += flightSkip
              djnz      flight, #:floop                    'while --flight <> 0

              shl       butterflySpan, #1                  'butterflySpan <<= 1
              shl       flightSkip, #1                     'flightSkip <<= 1

              shr       flight_max, #1                     'flight_max >>= 1
              shr       wSkip, #1                          'wSkip >>= 1
              shl       butterfly_max, #1                  'butterfly_max <<= 1
              djnz      level, #:lloop                     'while --level <> 0
:no_butterfly
              test      command, #CMD_MAGNITUDE wz         'Calculate magnitudes?
        if_z  jmp       #:no_magnitude

'Calculate magnitudes from the complex results in x and y. Results placed into x

              mov       c, fft_size_                       'repeat i from 0 to FFT_SIZE
              add       c, #1                              'That is one more than half FFT_SIZE so as
                                                           'to include the Nyquist frequency
              mov       b0x_ptr, bx_ptr
              mov       b0y_ptr, by_ptr

:mloop        rdlong    m1, b0x_ptr
              sar       m1, #LOG2_FFT_SIZE - 1
              mov       m2, m1
              call      #mul
              mov       input, m1

              rdlong    m1, b0y_ptr
              sar       m1, #LOG2_FFT_SIZE - 1
              mov       m2, m1
              call      #mul
              add       input, m1

              call      #sqrt

              wrlong    root, b0x_ptr                      'Write result to x array

              add       b0x_ptr, #4                        'Next x and y element and loop
              add       b0y_ptr, #4
              djnz      c, #:mloop

:no_magnitude
              mov       command, #0
              wrlong    command, mb_ptr
              jmp       #bfly
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
mul           'Account for sign
#ifdef USE_FASTER_MULT
              abs       m1, m1 wc
              negc      m2, m2
              abs       m2, m2 wc
              'Make t2 the smaller of the 2 unsigned parameters
              mov       m3, m1
              max       m3, m2
              min       m2, m1
              'Correct the sign of the adder
              negc      m2, m2
#else
              abs       m3, m1 wc
              negc      m2, m2
#endif
              'My accumulator
              mov       m1, #0
              'Do the work
:mul_loop     shr       m3, #1 wc,wz                       'Get the low bit of t2
        if_c  add       m1, m2                             'If it was a 1, add adder to accumulator
              shl       m2, #1                             'Shift the adder left by 1 bit
        if_nz jmp       #:mul_loop                         'Continue as long as there are no more 1's
mul_ret       ret

m1            long      0
m2            long      0
m3            long      0
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
#ifdef USE_FASTER_SQRT
'Faster code for square root (Chip Gracey after discussion with lonesock on Propeller Forums):
sqrt          mov       root, h40000000
              cmpsub    input, root  wc
              sumnc     root, h40000000
              shr       root, #1

              or        root, h10000000
              cmpsub    input, root  wc
              sumnc     root, h10000000
              shr       root, #1

              or        root, h04000000
              cmpsub    input, root  wc
              sumnc     root, h04000000
              shr       root, #1

              or        root, h01000000
              cmpsub    input, root  wc
              sumnc     root, h01000000
              shr       root, #1

              or        root, h00400000
              cmpsub    input, root  wc
              sumnc     root, h00400000
              shr       root, #1

              or        root, h00100000
              cmpsub    input, root  wc
              sumnc     root, h00100000
              shr       root, #1

              or        root, h00040000
              cmpsub    input, root  wc
              sumnc     root, h00040000
              shr       root, #1

              or        root, h00010000
              cmpsub    input, root  wc
              sumnc     root, h00010000
              shr       root, #1

              or        root, h00004000
              cmpsub    input, root  wc
              sumnc     root, h00004000
              shr       root, #1

              or        root, h00001000
              cmpsub    input, root  wc
              sumnc     root, h00001000
              shr       root, #1

              or        root, h00000400
              cmpsub    input, root  wc
              sumnc     root, h00000400
              shr       root, #1

              or        root, #$100
              cmpsub    input, root  wc
              sumnc     root, #$100
              shr       root, #1

              or        root, #$40
              cmpsub    input,root  wc
              sumnc     root, #$40
              shr       root, #1

              or        root, #$10
              cmpsub    input,root  wc
              sumnc     root, #$10
              shr       root, #1

              or        root, #$4
              cmpsub    input,root  wc
              sumnc     root, #$4
              shr       root, #1

              or        root, #$1
              cmpsub    input,root  wc
              sumnc     root, #$1
              shr       root, #1
sqrt_ret      ret

h10000000     long      $10000000
h04000000     long      $04000000
h01000000     long      $01000000
h00400000     long      $00400000
h00100000     long      $00100000
h00040000     long      $00040000
h00010000     long      $00010000
h00004000     long      $00004000
h00001000     long      $00001000
h00000400     long      $00000400

#else

'Faster code for square root (Chip Gracey after discussion with lonesock on Propeller Forums):
sqrt          mov       root, #0                           'Reset root
              mov       mask, h40000000                    'Reset mask (constant in register)
:sqloop       or        root, mask                         'Set trial bit
              cmpsub    input, root wc                     'Subtract root from input if fits
              sumnc     root, mask                         'Cancel trial bit, set root bit if fit
              shr       root, #1                           'Shift root down
              shr       mask, #2                           'Shift mask down
              tjnz      mask, #:sqloop                     'Loop until mask empty
sqrt_ret      ret
#endif
h40000000     long      $40000000
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
' Get sine/cosine
'
'      quadrant: 1              2             3             4
'          angle: $0000..$07FF $0800..$0FFF $1000..$17FF $1800..$1FFF
' table index: $0000..$07FF $0800..$0001 $0000..$07FF $0800..$0001
'        mirror: +offset        -offset       +offset       -offset
'           flip: +sample       +sample       -sample       -sample
'
' On entry: sin[12..0] holds angle (0° to just under 360°)
' On exit: sin holds signed value ranging from $0000FFFF ('1') to
' $FFFF0001 ('-1')
'
getcos        add       sin, sin_90                        'For cosine, add 90°
getsin        test      sin, sin_90      wc                'Get quadrant 2|4 into c
              test      sin, sin_180     wz                'Get quadrant 3|4 into nz
              negc      sin, sin                           'If quadrant 2|4, negate offset
              or        sin, sin_table                     'OR in sin table address >> 1
              shl       sin, #1                            'Shift left to get final word address
              rdword    sin, sin                           'Read word sample from $E000 to $F000
              negnz     sin, sin                           'If quadrant 3|4, negate sample
getsin_ret
getcos_ret    ret                                          '39..54 clocks
                                                           '(variance due to HUB sync on RDWORD)
sin_90        long      $0800
sin_180       long      $1000
sin_table     long      $E000 >> 1                         'ROM sin table base shifted right
sin           long      0
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
'Large constants
fft_size_     long      FFT_SIZE

'COG variables
level         long      0
flight        long      0
butterfly     long      0
flight_max    long      0
wSkip         long      0
butterflySpan long      0
butterfly_max long      0
flightSkip    long      0
k1            long      0
k2            long      0
k3            long      0
a             long      0
b             long      0
c             long      0
d             long      0
tx            long      0
ty            long      0
wx_ptr_       long      0
wy_ptr_       long      0
b0x_ptr       long      0
b0y_ptr       long      0
b1x_ptr       long      0
b1y_ptr       long      0
mb_ptr        long      0
bx_ptr        long      0
by_ptr        long      0

rev_a         long      0
hub_ptr       long      0
hub_rev_ptr   long      0
command       long      0
root          long      0
mask          long      0
input         long      0

wx_ptr        long      0-0
wy_ptr        long      0-0

'----------------------------------------------------------------------------------------------------------------------
              fit       496
#endif
'----------------------------------------------------------------------------------------------------------------------
DAT
'The W "twiddle factor" tables.
'These are one half cycle of cos and minus sine respectively scaled up to +/- 4096
wx  word  4096,  4095,  4095,  4095,  4094,  4094,  4093,  4092,  4091,  4089,  4088,  4086,  4084,  4082,  4080,  4078
    word  4076,  4073,  4071,  4068,  4065,  4062,  4058,  4055,  4051,  4047,  4043,  4039,  4035,  4031,  4026,  4022
    word  4017,  4012,  4007,  4001,  3996,  3990,  3985,  3979,  3973,  3967,  3960,  3954,  3947,  3940,  3933,  3926
    word  3919,  3912,  3904,  3897,  3889,  3881,  3873,  3864,  3856,  3848,  3839,  3830,  3821,  3812,  3803,  3793
    word  3784,  3774,  3764,  3754,  3744,  3734,  3723,  3713,  3702,  3691,  3680,  3669,  3658,  3647,  3635,  3624
    word  3612,  3600,  3588,  3576,  3563,  3551,  3538,  3526,  3513,  3500,  3487,  3473,  3460,  3447,  3433,  3419
    word  3405,  3391,  3377,  3363,  3348,  3334,  3319,  3304,  3289,  3274,  3259,  3244,  3229,  3213,  3197,  3182
    word  3166,  3150,  3134,  3117,  3101,  3085,  3068,  3051,  3034,  3018,  3000,  2983,  2966,  2949,  2931,  2914
    word  2896,  2878,  2860,  2842,  2824,  2806,  2787,  2769,  2750,  2732,  2713,  2694,  2675,  2656,  2637,  2617
    word  2598,  2578,  2559,  2539,  2519,  2500,  2480,  2460,  2439,  2419,  2399,  2379,  2358,  2337,  2317,  2296
    word  2275,  2254,  2233,  2212,  2191,  2170,  2148,  2127,  2105,  2084,  2062,  2040,  2018,  1997,  1975,  1952
    word  1930,  1908,  1886,  1864,  1841,  1819,  1796,  1773,  1751,  1728,  1705,  1682,  1659,  1636,  1613,  1590
    word  1567,  1544,  1520,  1497,  1474,  1450,  1427,  1403,  1379,  1356,  1332,  1308,  1284,  1260,  1237,  1213
    word  1189,  1164,  1140,  1116,  1092,  1068,  1043,  1019,   995,   970,   946,   921,   897,   872,   848,   823
    word   799,   774,   749,   725,   700,   675,   650,   625,   601,   576,   551,   526,   501,   476,   451,   426
    word   401,   376,   351,   326,   301,   276,   251,   226,   200,   175,   150,   125,   100,    75,    50,    25
wy  word     0,   -25,   -50,   -75,  -100,  -125,  -150,  -175,  -200,  -226,  -251,  -276,  -301,  -326,  -351,  -376
    word  -401,  -426,  -451,  -476,  -501,  -526,  -551,  -576,  -601,  -625,  -650,  -675,  -700,  -725,  -749,  -774
    word  -799,  -823,  -848,  -872,  -897,  -921,  -946,  -970,  -995, -1019, -1043, -1068, -1092, -1116, -1140, -1164
    word -1189, -1213, -1237, -1260, -1284, -1308, -1332, -1356, -1379, -1403, -1427, -1450, -1474, -1497, -1520, -1544
    word -1567, -1590, -1613, -1636, -1659, -1682, -1705, -1728, -1751, -1773, -1796, -1819, -1841, -1864, -1886, -1908
    word -1930, -1952, -1975, -1997, -2018, -2040, -2062, -2084, -2105, -2127, -2148, -2170, -2191, -2212, -2233, -2254
    word -2275, -2296, -2317, -2337, -2358, -2379, -2399, -2419, -2439, -2460, -2480, -2500, -2519, -2539, -2559, -2578
    word -2598, -2617, -2637, -2656, -2675, -2694, -2713, -2732, -2750, -2769, -2787, -2806, -2824, -2842, -2860, -2878
    word -2896, -2914, -2931, -2949, -2966, -2983, -3000, -3018, -3034, -3051, -3068, -3085, -3101, -3117, -3134, -3150
    word -3166, -3182, -3197, -3213, -3229, -3244, -3259, -3274, -3289, -3304, -3319, -3334, -3348, -3363, -3377, -3391
    word -3405, -3419, -3433, -3447, -3460, -3473, -3487, -3500, -3513, -3526, -3538, -3551, -3563, -3576, -3588, -3600
    word -3612, -3624, -3635, -3647, -3658, -3669, -3680, -3691, -3702, -3713, -3723, -3734, -3744, -3754, -3764, -3774
    word -3784, -3793, -3803, -3812, -3821, -3830, -3839, -3848, -3856, -3864, -3873, -3881, -3889, -3897, -3904, -3912
    word -3919, -3926, -3933, -3940, -3947, -3954, -3960, -3967, -3973, -3979, -3985, -3990, -3996, -4001, -4007, -4012
    word -4017, -4022, -4026, -4031, -4035, -4039, -4043, -4047, -4051, -4055, -4058, -4062, -4065, -4068, -4071, -4073
    word -4076, -4078, -4080, -4082, -4084, -4086, -4088, -4089, -4091, -4092, -4093, -4094, -4094, -4095, -4095, -4095
    word -4096, -4095, -4095, -4095, -4094, -4094, -4093, -4092, -4091, -4089, -4088, -4086, -4084, -4082, -4080, -4078
    word -4076, -4073, -4071, -4068, -4065, -4062, -4058, -4055, -4051, -4047, -4043, -4039, -4035, -4031, -4026, -4022
    word -4017, -4012, -4007, -4001, -3996, -3990, -3985, -3979, -3973, -3967, -3960, -3954, -3947, -3940, -3933, -3926
    word -3919, -3912, -3904, -3897, -3889, -3881, -3873, -3864, -3856, -3848, -3839, -3830, -3821, -3812, -3803, -3793
    word -3784, -3774, -3764, -3754, -3744, -3734, -3723, -3713, -3702, -3691, -3680, -3669, -3658, -3647, -3635, -3624
    word -3612, -3600, -3588, -3576, -3563, -3551, -3538, -3526, -3513, -3500, -3487, -3473, -3460, -3447, -3433, -3419
    word -3405, -3391, -3377, -3363, -3348, -3334, -3319, -3304, -3289, -3274, -3259, -3244, -3229, -3213, -3197, -3182
    word -3166, -3150, -3134, -3117, -3101, -3085, -3068, -3051, -3034, -3018, -3000, -2983, -2966, -2949, -2931, -2914
    word -2896, -2878, -2860, -2842, -2824, -2806, -2787, -2769, -2750, -2732, -2713, -2694, -2675, -2656, -2637, -2617
    word -2598, -2578, -2559, -2539, -2519, -2500, -2480, -2460, -2439, -2419, -2399, -2379, -2358, -2337, -2317, -2296
    word -2275, -2254, -2233, -2212, -2191, -2170, -2148, -2127, -2105, -2084, -2062, -2040, -2018, -1997, -1975, -1952
    word -1930, -1908, -1886, -1864, -1841, -1819, -1796, -1773, -1751, -1728, -1705, -1682, -1659, -1636, -1613, -1590
    word -1567, -1544, -1520, -1497, -1474, -1450, -1427, -1403, -1379, -1356, -1332, -1308, -1284, -1260, -1237, -1213
    word -1189, -1164, -1140, -1116, -1092, -1068, -1043, -1019,  -995,  -970,  -946,  -921,  -897,  -872,  -848,  -823
    word  -799,  -774,  -749,  -725,  -700,  -675,  -650,  -625,  -601,  -576,  -551,  -526,  -501,  -476,  -451,  -426
    word  -401,  -376,  -351,  -326,  -301,  -276,  -251,  -226,  -200,  -175,  -150,  -125,  -100,   -75,   -50,   -25
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
'    This file is distributed under the terms of the The MIT License as follows:
'
'    Copyright (c) 2011 Michael Rychlik
'
'    Permission is hereby granted, free of charge, to any person obtaining a copy
'    of this software and associated documentation files (the "Software"), to deal
'    in the Software without restriction, including without limitation the rights
'    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
'    copies of the Software, and to permit persons to whom the Software is
'    furnished to do so, subject to the following conditions:

'    The above copyright notice and this permission notice shall be included in
'    all copies or substantial portions of the Software.

'    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
'    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
'    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
'    AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
'    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
'    OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
'    THE SOFTWARE.
'----------------------------------------------------------------------------------------------------------------------

'----------------------------------------------------------------------------------------------------------------------
'The end.
